from blinker import signal
from . import emit_push_event, emit_user_push_event
from ..utils import unknown_value


exposed_signals = {}
exposed_signal_emitted = signal('exposed_signal_emitted')


def expose_signal(signal, marshaller=None, filter=None, room_getter=None, push_event_name=None, **kwargs):
    if not push_event_name:
        push_event_name = signal.name

    def marshal(sender, **kw):
        if marshaller:
            return marshaller(sender, **kw)
        return kw or None

    def emit(sender, **kw):
        emit_kwargs = dict(kwargs)
        if '_push_skip_self' in kw:
            emit_kwargs['skip_self'] = kw.pop('_push_skip_self')
        data = marshal(sender, **kw)
        if room_getter:
            room, is_user = room_getter(sender)
        else:
            room, is_user = getattr(sender, 'exposed_signal_room', (None, False))
        if is_user:
            emit_user_push_event(room, push_event_name, data, **emit_kwargs)
        else:
            emit_push_event(push_event_name, data, room=room, **emit_kwargs)
        return data

    def listener(sender, **kw):
        skip_expose = kw.pop('_skip_expose', False)
        skip_push = kw.pop('_skip_push', skip_expose)
        skip_self =  kw.pop('_push_skip_self', True)
        if filter and not filter(sender, **kw):
            return
        push_data = unknown_value
        if not skip_push:
            push_data = emit(sender, _push_skip_self=skip_self, **kw)
        if not skip_expose:
            if push_data is unknown_value:
                push_data = marshal(sender, **kw)
            exposed_signal_emitted.send(sender, signal_name=signal.name, signal_data=kw, push_data=push_data)

    listener.marshal = marshal
    listener.emit = emit
    listener.filter = filter
    signal.exposed = listener
    exposed_signals[signal] = listener # we make sure the reference to listener is kept so we don't use a weak listener
    signal.connect(listener)
    return signal


def exposed_signal(name, **kwargs):
    return expose_signal(signal(name, doc=kwargs.pop('doc', None)), **kwargs)



from flask import g
from frasco.ext import *


class FrascoSubdomains(Extension):
    name = "frasco_subdomains"
    defaults = {"param_name": "subdomain"}

    def _init_app(self, app, state):
        @app.url_value_preprocessor
        def extract_subdomain_from_values(endpoint, values):
            if values:
                g.subdomain = values.pop(state.options["param_name"], None)
        
        @app.url_defaults
        def add_subdomain_to_url_params(endpoint, values):
            if state.options["param_name"] not in values:
                values[state.options["param_name"]] = g.subdomain



from frasco.ext import *
from frasco.i18n import lazy_translate
from frasco.utils import populate_obj, extract_unmatched_items
from flask import render_template, abort, request
from flask_login import LoginManager, logout_user, login_required, login_url, login_fresh, confirm_login, fresh_login_required, user_logged_in
import datetime
import os

from .user import *
from .model import *
from .jinja_ext import *
from .forms import *
from .tokens import *
from .tokens import TOKEN_NS_ACCESS_TOKEN
from .signals import *
from .password import *
from .blueprint import users_blueprint
from . import captcha


class FrascoUsersState(ExtensionState):
    def __init__(self, *args, **kwargs):
        super(FrascoUsersState, self).__init__(*args, **kwargs)
        self.manager = LoginManager()
        self.user_validators = []
        self.override_builtin_user_validation = False
        self.login_validators = []
        self.password_validators = []
        self.captcha_validator = None
        self.user_request_loaders = []

    def user_request_loader(self, func):
        self.user_request_loaders.append(func)
        return func


class FrascoUsers(Extension):
    name = "frasco_users"
    state_class = FrascoUsersState
    defaults = {
        # email
        "must_provide_email": True,
        "email_is_unique": True,
        "email_allowed_domains": None,
        # username
        "must_provide_username": True,
        "username_is_unique": True,
        "forbidden_usernames": [],
        "min_username_length": 1,
        "allow_spaces_in_username": False,
        "username_case_sensitive": False,
        # password
        "validate_password_regexps": None,
        "prevent_password_reuse": False,
        "max_password_reuse_saved": None,
        "min_time_between_password_change": None,
        "expire_password_after": None,
        "max_password_length": 500, # used to prevent DOS attacks
        # login
        "allow_login": True,
        "enable_2fa": False,
        "login_view": "users.login",
        "login_redirect": None, # redirect to url instead of login page
        "login_form_class": LoginWithEmailForm,
        "login_2fa_form_class": Login2FAForm,
        "allow_email_or_username_login": True,
        "remember_days": 365,
        "redirect_after_login": "index",
        "redirect_after_login_disallowed": None,
        "2fa_issuer_name": None, # default is app.config['TITLE']
        "2fa_remember_days": 60,
        "2fa_remember_cookie_options": {},
        # signup
        "signup_redirect": None, # redirect to url instead of signup page
        "allow_signup": True,
        "signup_form_class": SignupForm,
        "send_welcome_email": False,
        "login_user_on_signup": True,
        "rate_limit_count": None,
        "rate_limit_period": 60,
        "redirect_after_signup": "index",
        "redirect_after_signup_disallowed": None, # go to login
        # captcha
        "debug_captcha": False,
        "recaptcha_key": None,
        "recaptcha_secret": None,
        "hcaptcha_key": None,
        "hcaptcha_secret": None,
        # reset password
        "reset_password_redirect": None, # redirect to url instead of reset password page
        "allow_reset_password": True,
        "send_reset_password_form_class": SendResetPasswordForm,
        "reset_password_form_class": ResetPasswordForm,
        "send_reset_password_email": True,
        "reset_password_ttl": 3600, # 1h
        "login_user_on_reset_password": True,
        "redirect_after_reset_password_token": False,
        "redirect_after_reset_password": "index",
        "redirect_after_reset_password_disallowed": "users.login",
        # email validation
        "redirect_after_email_validated": "index",
        "email_validation_ttl": None,
        "block_non_email_validated_users": False,
        "send_email_validation_email": False,
        "login_after_email_validation": True,
        "allow_non_email_validated_users_who_logged_in_before": True,
        # logout
        "redirect_after_logout": "index",
        # oauth
        "oauth_signup_only": False,
        "oauth_login_only": False,
        "oauth_must_signup": False,
        "oauth_must_provide_password": False,
        # auth
        "disable_password_authentication": False,
        "default_auth_provider_name": "app",
        # access tokens
        "enable_access_tokens": False,
        "access_tokens_ttl": None,
        "enable_access_tokens_web_flow": True,
        "access_tokens_web_flow_allowed_redirects": [],
        # messages
        "login_error_message": lazy_translate("Invalid email or password"),
        "login_disallowed_message": None,
        "login_2fa_error_message": lazy_translate("Invalid two factor authentification code"),
        "login_required_message": lazy_translate("Please log in to access this page"),
        "fresh_login_required_message": lazy_translate("Please reauthenticate to access this page"),
        "password_expired_message": lazy_translate("Your password has expired, please enter a new one"),
        "must_provide_username_message": lazy_translate("A username must be provided"),
        "password_reused_message": lazy_translate("You cannot use a password which you have previously used"),
        "min_time_between_password_change_message": lazy_translate("You have changed your password too recently"),
        "validate_password_regexps_message": lazy_translate("The password does not respect the following rule: {rule}"),
        "max_password_length_message": lazy_translate("The password is too long"),
        "must_provide_email_message": lazy_translate("An email address must be provided"),
        "signup_disallowed_message": None,
        "username_taken_message": lazy_translate("An account using the same username already exists"),
        "email_taken_message": lazy_translate("An account using the same email already exists"),
        "username_too_short_message": lazy_translate("The username is too short"),
        "username_has_spaces_message": lazy_translate("The username cannot contain spaces"),
        "password_confirm_failed_message": lazy_translate("The two passwords do not match"),
        "bad_signup_code_message": lazy_translate("The provided code is not valid"),
        "rate_limit_reached_message": lazy_translate("Too many accounts have been created from this location in a too short period. Please, try again later"),
        "reset_password_token_error_message": lazy_translate("This email does not exist in our database"),
        "reset_password_token_success_message": lazy_translate("An email has been sent to your email address with a link to reset your password"),
        "reset_password_error_message": lazy_translate("Invalid or expired link to reset your password"),
        "reset_password_success_message": lazy_translate("Password successfully resetted"),
        "reset_password_disallowed_message": lazy_translate("You are not allowed to reset your password"),
        "update_password_error_message": lazy_translate("Invalid current password"),
        "update_user_email_error_message": lazy_translate("An account using the same email already exists"),
        "oauth_user_denied_login": lazy_translate("Login was denied"),
        "oauth_user_already_exists_message": lazy_translate("This {provider} account has already been used on a different account"),
        "oauth_error": lazy_translate("An error occured while authentifying you with the remote provider"),
        "captcha_fail_message": lazy_translate("The captcha validation has failed"),
        "email_validation_success_message": None,
        "enable_admin": True
    }

    def _init_app(self, app, state):
        state.Model = state.import_option('model')
        state.LoginModel = state.import_option('login_model', required=False)
        
        app.config.setdefault("REMEMBER_COOKIE_DURATION", datetime.timedelta(days=state.options["remember_days"]))

        app.register_blueprint(users_blueprint)
        app.jinja_env.add_extension(LoginRequiredExtension)
        app.jinja_env.add_extension(AnonymousOnlyExtension)
        
        state.manager.init_app(app)
        state.manager.login_view = state.options['login_view']
        state.manager.login_message_category = "error"
        populate_obj(state.manager, extract_unmatched_items(state.options, self.defaults))

        if state.options['recaptcha_key']:
            state.captcha_validator = captcha.validate_recaptcha
        elif state.options['hcaptcha_key']:
            state.captcha_validator = captcha.validate_hcaptcha

        if has_extension("frasco_mail", app):
            app.extensions.frasco_mail.add_templates_from_package(__name__)
        if has_extension("frasco_babel", app):
            app.extensions.frasco_babel.add_extract_dir(os.path.dirname(__file__), ["templates"])

        @state.manager.user_loader
        def user_loader(id):
            return state.Model.query.filter(getattr(state.Model, getattr(state.Model, '__session_cookie_identifier__', 'id')) == id).first()

        @state.manager.request_loader
        def request_loaders(request):
            for loader in state.user_request_loaders:
                user = loader(request)
                if user:
                    return user
        
        if state.options['enable_access_tokens']:
            @state.user_request_loader
            def access_token_user_loader(request):
                access_token = request.args.get('access_token')
                if 'Authorization' in request.headers:
                    authz = request.headers['Authorization']
                    if authz.startswith('Bearer '):
                        access_token = authz[7:]
                if access_token:
                    return read_user_token(access_token, TOKEN_NS_ACCESS_TOKEN, state.options['access_tokens_ttl'])

        if state.options['block_non_email_validated_users']:
            @app.before_request
            def block_non_email_validated_users():
                if current_user.is_authenticated and current_user.must_validate_email and (not request.endpoint or \
                  (request.endpoint not in ('users.logout', 'users.send_email_validation_email', 'users.validate_email', 'static') and \
                  not request.endpoint.endswith('.static'))):
                    if request.is_json:
                        abort(403)
                    return render_template("users/non_email_validated_users_block_page.html", email=current_user.email)

    @ext_stateful_method
    def user_validator(self, state, func):
        state.user_validators.append(func)
        return func

    @ext_stateful_method
    def login_validator(self, state, func):
        state.login_validators.append(func)
        return func

    @ext_stateful_method
    def password_validator(self, state, func):
        state.password_validators.append(func)
        return func



from flask import request
from frasco.ext import get_extension_state
import requests


def validate_recaptcha():
    state = get_extension_state('frasco_users')
    if 'g-recaptcha-response' in request.form:
        r = requests.post('https://www.google.com/recaptcha/api/siteverify', data={
            'secret': state.options['recaptcha_secret'],
            'response': request.form['g-recaptcha-response'],
            'remote_ip': request.remote_addr
        })
        return r.ok and r.json().get('success')
    return False


def validate_hcaptcha():
    state = get_extension_state('frasco_users')
    if 'h-captcha-response' in request.form:
        r = requests.post('https://hcaptcha.com/siteverify', data={
            'secret': state.options['hcaptcha_secret'],
            'response': request.form['h-captcha-response'],
            'remoteip': request.remote_addr
        })
        return r.ok and r.json().get('success')
    return False



from frasco.ext import *
from frasco.mail import send_mail
from frasco.models import db, transaction
from flask.signals import Namespace as SignalNamespace
import datetime
import click
from contextlib import contextmanager
from .model import *


_signals = SignalNamespace()
invoice_issued = _signals.signal('invoice_issued')


class FrascoInvoicing(Extension):
    name = "frasco_invoicing"
    defaults = {"send_email": None}

    def _init_app(self, app, state):
        state.Model = state.import_option('model')
        state.ItemModel = state.import_option('item_model')
        state.ref_creator_func = create_invoice_ref

        if has_extension("frasco_mail", app):
            app.extensions.frasco_mail.add_templates_from_package(__name__)
            if state.options['send_email'] is None:
                state.options['send_email'] = True

        @app.cli.command('send-invoice-email')
        @click.argument('invoice_id')
        @click.option('--email')
        def send_email_command(invoice_id, email=None):
            invoice = state.Model.query.get(invoice_id)
            send_invoice_mail(email or invoice.email, invoice)

    @ext_stateful_method
    def ref_creator(self, state, func):
        state.ref_creator_func = func
        return func


@contextmanager
def create_invoice(**create_invoice_ref_kwargs):
    state = get_extension_state('frasco_invoicing')
    with transaction():
        invoice = state.Model()
        invoice.ref = create_invoice_ref(**create_invoice_ref_kwargs)
        yield (invoice, state.ItemModel)
        db.session.add(invoice)
        invoice_issued.send(invoice)
    if state.options['send_email'] and invoice.email:
        send_invoice_mail(invoice.email, invoice)


def create_invoice_ref(category=None, counter=None, separator='-', merge_date=True):
    today = datetime.date.today()
    parts = [today.year, today.month, today.day]
    if merge_date:
        parts = ["".join(map(str, parts))]
    if category:
        parts.append(category)
    if counter is None:
        counter = get_extension_state('frasco_invoicing').Model.query.count() + 1
    parts.append(counter)
    return separator.join(map(str, parts))


def send_invoice_mail(email, invoice, **kwargs):
    items = []
    for item in invoice.items:
        items.append((item.description, item.quantity, item.amount))
    send_mail(email, 'invoice.html',
        invoice=invoice,
        invoice_date=invoice.issued_at,
        invoice_items=items,
        invoice_currency=invoice.currency.upper(),
        invoice_total=invoice.total,
        invoice_tax=invoice.tax_amount,
        invoice_tax_rate=invoice.tax_rate,
        **kwargs)


def send_failed_invoice_mail(email, invoice, **kwargs):
    items = []
    for line in invoice.lines.data:
        items.append((line.description or '', line.quantity, line.amount / 100.0))
    send_mail(email, 'failed_invoice.html',
        invoice_date=datetime.datetime.fromtimestamp(invoice.date),
        invoice_items=items,
        invoice_currency=invoice.currency.upper(),
        invoice_total=invoice.total / 100.0, **kwargs)



from frasco.ext import *
from frasco.upload import url_for_upload
from frasco.helpers import url_for
from frasco.utils import slugify
from flask import current_app, request
import sqlalchemy as sqla
import hashlib
import urllib.request, urllib.parse, urllib.error
import base64
import requests

from .first_letter_avatar import generate_first_letter_avatar_svg


def svg_to_base64_data(svg):
    return 'data:image/svg+xml;base64,' + base64.b64encode(svg)


class UserAvatarModelMixin(object):
    avatar_filename = sqla.Column(sqla.String)

    @property
    def avatar_url(self):
        return url_for_avatar(self)


class FrascoUsersAvatars(Extension):
    name = "frasco_users_avatars"
    defaults = {"url": None,
                "avatar_size": 80,
                "add_flavatar_route": False,
                "email_column": None,
                "name_column": None,
                "try_gravatar": True,
                "force_gravatar": False,
                "gravatar_size": None,
                "gravatar_default": "mm",
                "force_flavatar": False,
                "flavatar_size": "100%",
                "flavatar_font_size": 70,
                "flavatar_text_dy": "20%",
                "flavatar_length": 1,
                "flavatar_colors": ["#0E7C7B", "#17BEBB", "#D4F4DD", "#D62246", "#4B1D3F", "#C0DA74", "#BEEDAA", "#B2DDF7", "#FFFF82", "#13315C", "#6761A8", "#407887", "#0CCE6B", "#F1AB86", "#EAD94C", "#23B5D3", "#324376"]}

    def _init_app(self, app, state):
        app.add_template_global(url_for_avatar)

        def flavatar(name, bgcolorstr=None):
            if bgcolorstr is None:
                bgcolorstr = request.args.get('bgcolorstr')
            svg = generate_first_letter_avatar_svg(name, bgcolorstr, request.args.get('size'))
            return svg, 200, {
                'Content-Type': 'image/svg+xml',
                'Cache-Control': 'public, max-age=31536000'
            }

        @app.route('/avatar/<hash>/<name>')
        def avatar(hash, name):
            if state.options['try_gravatar']:
                size = state.options['gravatar_size'] or state.options["avatar_size"]
                try:
                    r = requests.get(url_for_gravatar(hash, size=size, default=404))
                    if r.status_code != 404:
                        return r.content, 200, {'Content-Type': r.headers['content-type']}
                except Exception:
                    pass
            return flavatar(name, hash)

        if state.options['add_flavatar_route']:
            app.add_url_rule('/flavatar/<name>.svg', 'flavatar', flavatar)
            app.add_url_rule('/flavatar/<name>/<bgcolorstr>.svg', 'flavatar', flavatar)


def url_for_avatar(user):
    state = get_extension_state('frasco_users_avatars')
    if getattr(user, 'avatar_filename', None):
        return url_for_upload(user.avatar_filename)

    hash = None
    username = getattr(user, state.options["name_column"] or 'username', None)
    if username:
        username = slugify(username.lower())
        hash = hashlib.md5(username.encode('utf-8')).hexdigest()

    email = getattr(user, state.options["email_column"] or 'email', None)
    if email:
        hash = hashlib.md5(email.lower().encode('utf-8')).hexdigest()
        if not username:
            username = slugify(email.split('@')[0])

    if state.options["force_flavatar"] and username:
        if state.options['add_flavatar_route']:
            return url_for('flavatar', name=username, bgcolorstr=hash, _external=True)
        return svg_to_base64_data(generate_first_letter_avatar_svg(username, hash))
    if state.options["force_gravatar"] and email:
        return url_for_gravatar(email)
    if state.options['url'] and email:
        return state.options["url"].format(email=email, email_hash=hash, username=username)
    return url_for('avatar', hash=hash, name=username, _external=True)


def url_for_gravatar(email, size=None, default=None):
    state = get_extension_state('frasco_users_avatars')
    hash = hashlib.md5(email.lower().encode('utf-8')).hexdigest()
    params = {
        's': size or state.options['gravatar_size'] or state.options["avatar_size"],
        'd': default or state.options['gravatar_default']
    }
    return "https://www.gravatar.com/avatar/%s?%s" % (hash, urllib.parse.urlencode({k: v for k, v in params.items() if v is not None}))

